use std::{
	io::{self, Write},
	path::PathBuf,
	str::FromStr,
};

use anyhow::{Result, anyhow};

use rusqlite::{Connection, params, params_from_iter};

use crate::{
	cli::{self, FilesArgs, TagArgs, TagsArgs, UntagArgs},
	try_database,
};

pub fn handle_files(args: FilesArgs) -> Result<()> {
	let conn = try_database()?;

	let mut sql = r#"SELECT path, group_concat(name) as tags FROM file_tag
			JOIN file ON file.id = file_tag.file_id
			JOIN tag ON tag.id = file_tag.tag_id"#
		.to_string();

	let mut show_tags = true;

	if let Some(query) = args.query {
		let path = PathBuf::from_str(&query).unwrap_or_default();
		match path.exists() {
			true => {
				sql.push_str(&format!(" WHERE path IN ('{}')", query));
			}
			false => {
				sql.push_str(&format!(" WHERE tag.name LIKE '%{}%'", query));
				show_tags = false;
			}
		};
	}

	sql.push_str(" GROUP BY path");

	let mut stmt = conn.prepare(&sql)?;
	let result: Vec<(String, String)> = stmt
		.query_map([], |row| Ok((row.get(0)?, row.get(1)?)))?
		.flatten()
		.collect();

	let mut w = io::stdout();
	if !result.is_empty() {
		for (file, tags) in result {
			if show_tags {
				writeln!(&mut w, "{}: {}", file, tags)?;
			} else {
				writeln!(&mut w, "{}", file)?;
			}
		}
	} else {
		writeln!(&mut w, "No files registered!")?;
	}
	w.flush()?;

	Ok(())
}

pub fn handle_tag(args: TagArgs) -> Result<()> {
	let mut conn = try_database()?;

	tag_file(&mut conn, args.file, args.tags)?;

	Ok(())
}

pub fn handle_untag(args: UntagArgs) -> Result<()> {
	let mut conn = try_database()?;

	untag_file(&mut conn, args.file, args.tags)?;

	Ok(())
}

pub fn handle_tags(args: TagsArgs) -> Result<()> {
	let mut conn = try_database()?;

	match args.commands {
		Some(cli::TagsCommands::List) | None => {
			let mut w = io::stdout();
			let tags = list_tags(&conn)?;
			if !tags.is_empty() {
				for tag in tags {
					writeln!(&mut w, "{}", tag)?;
				}
			} else {
				writeln!(&mut w, "No tags registered!")?;
			}
			w.flush()?;
			Ok(())
		}
		Some(cli::TagsCommands::Add { add }) => add_tags(&mut conn, add),
		Some(cli::TagsCommands::Remove { remove }) => remove_tags(&mut conn, remove),
		Some(cli::TagsCommands::Rename { old, new }) => rename_tag(&conn, old, new),
	}
}

fn rename_tag(conn: &Connection, old: String, new: String) -> Result<()> {
	let mut stmt = conn.prepare("UPDATE tag SET name = ? WHERE name = ?")?;

	stmt.execute([&new, &old])?;

	Ok(())
}

fn list_tags(conn: &Connection) -> Result<Vec<String>> {
	let mut stmt = conn.prepare("SELECT name FROM tag")?;
	let result = stmt.query_map([], |row| row.get(0))?.flatten();

	let mut tags = Vec::new();
	for name in result {
		tags.push(name);
	}

	Ok(tags)
}

fn remove_tags(conn: &mut Connection, tags: Vec<String>) -> Result<()> {
	let tx = conn.transaction()?;
	{
		let mut query = r#"DELETE FROM tag WHERE name IN ("#.to_string();

		for (i, _tag) in tags.iter().enumerate() {
			query.push('?');
			if i < tags.len() - 1 {
				query.push(',');
			}
		}

		query.push(')');

		tx.execute(&query, params_from_iter(&tags))?;

		if tx.changes() > 0 {
			let mut w = io::stdout();
			writeln!(&mut w, "Removed Tags:")?;
			for tag in tags {
				writeln!(&mut w, "  {tag}")?;
			}
		}
	}
	tx.commit()?;

	Ok(())
}

fn add_tags(conn: &mut Connection, tags: Vec<String>) -> Result<()> {
	let tx = conn.transaction()?;
	{
		let mut query = r#"INSERT INTO tag(name) VALUES"#.to_string();

		for (i, _tag) in tags.iter().enumerate() {
			query.push_str("(?)");
			if i < tags.len() - 1 {
				query.push(',');
			}
		}

		tx.execute(&query, params_from_iter(&tags))?;

		if tx.changes() > 0 {
			let mut w = io::stdout();
			writeln!(&mut w, "Added Tags:")?;
			for tag in tags {
				writeln!(&mut w, "  {tag}")?;
			}
		}
	}
	tx.commit()?;

	Ok(())
}

fn tag_file(conn: &mut Connection, file: PathBuf, tags: Vec<String>) -> Result<()> {
	let file = file
		.to_str()
		.ok_or(anyhow!("Unable to find path {:?}", file))?
		.to_string();

	let tx = conn.transaction()?;
	{
		let file_id: Option<i32> = tx
			.query_row(r#"SELECT id FROM file WHERE path = ?"#, [&file], |row| {
				row.get(0)
			})
			.ok();

		let file_id = match file_id {
			Some(id) => id,
			None => {
				let mut stmt = tx.prepare(r#"INSERT INTO file(path) VALUES (?) RETURNING id"#)?;
				stmt.query_row([file], |row| -> Result<i32, rusqlite::Error> { row.get(0) })?
			}
		};

		let mut sql = "SELECT id FROM tag WHERE name IN ".to_string();
		sql.push('(');
		let size = tags.len();
		for (i, _tag) in tags.iter().enumerate() {
			sql.push('?');
			if i < size - 1 {
				sql.push(',');
			}
		}
		sql.push(')');

		let mut stmt = tx.prepare(&sql)?;
		let result = stmt
			.query_map(
				params_from_iter(tags),
				|row| -> Result<i32, rusqlite::Error> { row.get(0) },
			)?
			.flatten();

		let mut tags_ids = Vec::new();
		for id in result {
			tags_ids.push(id);
		}

		let mut stmt = tx.prepare(r#"INSERT INTO file_tag(file_id, tag_id) VALUES (?, ?)"#)?;

		for tag_id in tags_ids {
			stmt.execute((file_id, tag_id))?;
		}
	}
	tx.commit()?;

	Ok(())
}

fn untag_file(conn: &mut Connection, file: PathBuf, tags: Option<Vec<String>>) -> Result<()> {
	let file = file
		.to_str()
		.ok_or(anyhow!("Unable to find path {:?}", file))?
		.to_string();

	let tx = conn.transaction()?;
	{
		let _file_id: i32 = tx
			.query_row(r#"SELECT id FROM file WHERE path = ?"#, [&file], |row| {
				row.get(0)
			})
			.map_err(|_| anyhow!("No {} file registered!", file))?;

		let mut sql = r#"
			DELETE FROM file_tag
			WHERE ROWID IN (
				SELECT file_tag.ROWID FROM file_tag
				INNER JOIN file ON file_tag.file_id = file.id
				INNER JOIN tag ON file_tag.tag_id = tag.id"#
			.to_string();

		match tags {
			Some(tags) => {
				let tags = tags.join(",");
				sql.push_str(" WHERE file.path = ?1 AND tag.name IN (?2))");
				let mut stmt = tx.prepare(&sql)?;
				stmt.execute(params![file, tags])?;
			}
			None => {
				sql.push_str(" WHERE file.path = ?1)");
				let mut stmt = tx.prepare(&sql)?;
				stmt.execute(params![file])?;
			}
		};
	}
	tx.commit()?;

	Ok(())
}
